/* 
   Figment - a Gtk2 based outliner.  

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "figment.h"
#include "../config.h"

#define _(_x) gettext (_x)
#define N_(_x) (_x)

static void menu_about_callback( GtkWidget *w, gpointer op);
static void menu_add_callback( GtkWidget *w, gpointer   data );
static void remove_item( GtkWidget *w, gpointer   data );

static void menu_view_callback( GtkWidget *w, gpointer op);
static void menu_export_callback( GtkWidget *w, gpointer   data );
static void menu_file_callback( GtkWidget *w, gpointer   data );
static void fig_exit( GtkWidget *w, gpointer   data );
static int  fig_question(gchar *text);
static void fig_options_dialog(GtkWidget *w, gpointer   data);
static void toggle_toolbar(GtkWidget *w, gpointer   data);

Figment Fig;

/* Branch copying code */

typedef struct branch_copy {
  gchar    *text;
  gboolean checked;
  gboolean expanded;
  struct branch_copy *next;
  struct branch_copy *child;
} BranchCopy;

BranchCopy*
branch_new(gchar *text, gboolean checked, gboolean expanded)
{
  BranchCopy *b = malloc(sizeof(struct branch_copy));
  b->text = text;
  b->checked = checked;
  b->expanded = expanded;
  b->next = NULL;
  b->child = NULL;
  return b;
}

enum {
  BRANCH_COPY_KIDS,
  BRANCH_COPY_ALL,
  BRANCH_COPY_ADD_AS_SIBLING_BEFORE,
  BRANCH_COPY_ADD_AS_SIBLING_AFTER,
  BRANCH_COPY_ADD,
  BRANCH_COPY_ADD_AS_CHILD,
};

static BranchCopy*
branch_recurse_copy(GtkTreeIter *iter, int mode)
{
  GtkTreeModel *model;
  gchar *text;
  GtkTreePath *path;
  GtkTreeIter child_iter;
  gboolean checked;
  gboolean expanded;
  gboolean valid = 1;

  BranchCopy *cur = NULL, *result = NULL;

  model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));
  while (valid)
    {
      gtk_tree_model_get (model, iter, 
                          THOUGHT_COLUMN, &text,
                          CHECKED_COLUMN, &checked,
                          -1);

      path = gtk_tree_model_get_path(model, iter);
      expanded = gtk_tree_view_row_expanded(GTK_TREE_VIEW(Fig.treeview), 
					    path);
      gtk_tree_path_free(path);
      
      if (cur == NULL)
	result = cur = branch_new(text, checked, expanded);
      else 
	{
	  cur->next = branch_new(text, checked, expanded);
	  cur = cur->next;
	}



      if (gtk_tree_model_iter_children(model, &child_iter, iter))
	cur->child = branch_recurse_copy(&child_iter, BRANCH_COPY_ALL);
      
      if (mode == BRANCH_COPY_KIDS) 
	valid = 0;
      else
	valid = gtk_tree_model_iter_next (model, iter);
    }

  return result;
}

void
branch_add(GtkTreeIter *iter_mount, BranchCopy *branch, 
	   gint mode, gboolean parent_expanded) 
{
  GtkTreeIter iter_new;
  GtkTreePath *path;
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));

  while(branch != NULL)
    {
      switch(mode)
	{
	case BRANCH_COPY_ADD_AS_SIBLING_AFTER: 
	  gtk_tree_store_insert_after(GTK_TREE_STORE(model),
				      &iter_new,
				      NULL,
				      iter_mount
				      );
	  break;
	case BRANCH_COPY_ADD_AS_SIBLING_BEFORE: 
	  gtk_tree_store_insert_before(GTK_TREE_STORE(model),
				      &iter_new,
				      NULL,
				      iter_mount
				       );
	  break;
	default : 		/* Added as child */
	  gtk_tree_store_insert_after(GTK_TREE_STORE(model),
				    &iter_new,
				    iter_mount,
				    NULL );
	}


      gtk_tree_store_set (GTK_TREE_STORE(model), &iter_new,
			  CHECKED_COLUMN, 
			  branch->checked,
			  EDITABLE_COLUMN, TRUE,
			  -1);

      figment_set_cell_text(&iter_new, branch->text);

      if (parent_expanded)
	{
	  path = gtk_tree_model_get_path(model, iter_mount);
	  gtk_tree_path_up(path);
	  gtk_tree_view_expand_row(GTK_TREE_VIEW(Fig.treeview), path, TRUE);
	  gtk_tree_path_free(path);
	}

      if (branch->child != NULL)
	branch_add(&iter_new, branch->child, BRANCH_COPY_ADD_AS_CHILD,
		   branch->expanded);

      branch = branch->next;
    }
}

void 
branch_free(BranchCopy *branch)
{
  /* TODO */
  return;
}

/* Menus */

static GtkItemFactoryEntry menu_items[] = {
  { N_("/_File"),         NULL,         NULL,           0, "<Branch>" },
  { N_("/File/_New"),     "<control>N", menu_file_callback,    MENU_OP_FILE_NEW, "<Item>" },
  { N_("/File/_Open"),    "<control>O", menu_file_callback,    MENU_OP_FILE_OPEN, "<Item>" },
  { N_("/File/_Save"),    "<control>S", menu_file_callback,    MENU_OP_FILE_SAVE, "<Item>" },
  { N_("/File/Save _As"), NULL, menu_file_callback, MENU_OP_FILE_SAVE_AS, "<Item>" },
  { N_("/File/Revert"), NULL, menu_file_callback, MENU_OP_FILE_REVERT, "<Item>" },
  { N_("/File/sep1"),     NULL,         NULL,           0, "<Separator>" },
  { N_("/File/Export"),   NULL,         NULL,    0, "<Branch>" },
  { N_("/File/Export/Plain Text"), NULL,  menu_export_callback, MENU_OP_EXPORT_TEXT, "<Item>" },
  { N_("/File/Export/HTML"),       NULL,  menu_export_callback,   MENU_OP_EXPORT_HTML, "<Item>" },
  /*  { "/File/Export/RTF",   NULL,    menu_export_callback,   MENU_OP_EXPORT_RTF, "<Item>" }, */
  { N_("/File/Export/DocBook"),NULL,   menu_export_callback,   MENU_OP_EXPORT_DOCBOOK, "<Item>" },
  { N_("/File/sep2"),     NULL,         NULL,           0, "<Separator>" },
  { N_("/File/Quit"),     "<control>Q", fig_exit,  0, "<Item>" },

  { N_("/_View"),         NULL,         NULL,           0, "<Branch>" },
  { N_("/View/Expand"), NULL,   menu_view_callback,  
    MENU_OP_EXPAND, "<Item>" },
  { N_("/View/Expand Subheads"), NULL, menu_view_callback, 
    MENU_OP_EXPAND_SUBS, "<Item>" },
  { N_("/View/Expand All"), NULL, menu_view_callback, 
    MENU_OP_EXPAND_ALL, "<Item>" },
  { N_("/View/sep2"),     NULL,            NULL,           0, "<Separator>" },
  { N_("/View/Collapse"), NULL, 
    menu_view_callback, MENU_OP_COLLAPSE, "<Item>" },
  { N_("/View/Collapse To Parent"), NULL, 
    menu_view_callback, MENU_OP_COLLAPSE_PAR, "<Item>" },
  { N_("/View/Collapse All"), NULL, menu_view_callback, 
    MENU_OP_COLLAPSE_ALL, "<Item>" },
  { N_("/View/sep3"),     NULL,            NULL,           0, "<Separator>" },
  { N_("/View/Toggle Toolbar"), NULL, toggle_toolbar, 0, "<Item>" },

  { N_("/_Outline"),      NULL,         NULL,           0, "<Branch>" },
  { N_("/Outline/Add Sibling"), "<alt>a", menu_add_callback,   MENU_OP_ADD_SIB, "<Item>" },
  { N_("/Outline/Add Child"),   "<alt>c", menu_add_callback,     MENU_OP_ADD_CHLD   ,   "<Item>" },
  { N_("/Outline/Remove"),"<alt>d", remove_item,       0, "<Item>" },
  { N_("/Outline/sep4"), NULL, NULL,       0, "<Separator>" },
  { N_("/Outline/Promote"), "<alt>p", menu_add_callback, MENU_OP_PROMOTE, "<Item>" },
  { N_("/Outline/Demote"), "<alt>d", menu_add_callback, MENU_OP_DEMOTE, "<Item>" },
  { N_("/Outline/Raise"), "<alt>r", menu_add_callback, MENU_OP_RAISE, "<Item>" },
  { N_("/Outline/Lower"), "<alt>l", menu_add_callback, MENU_OP_LOWER, "<Item>" },
  { N_("/Outline/sep5"), NULL, NULL,       0, "<Separator>" },
  { N_("/Outline/Options"), NULL, fig_options_dialog,   0, "<Item>" },
  { N_("/_Help"),         NULL,         NULL,           0, "<LastBranch>" },
  { N_("/_Help/About"),   NULL,         menu_about_callback,    0, "<Item>" },
};

static gint nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);

static void
menu_about_callback( GtkWidget *w, gpointer op)
{
  GtkWidget *msg;
  msg = gtk_message_dialog_new(GTK_WINDOW(Fig.window),
			       GTK_DIALOG_DESTROY_WITH_PARENT,
			       GTK_MESSAGE_INFO,
			       GTK_BUTTONS_OK,
			       "Figment\nVersion " VERSION "\nmallum@handhelds.org");

  g_signal_connect_swapped (GTK_OBJECT (msg), 
			    "response", 
			    G_CALLBACK (gtk_widget_destroy),
			    GTK_OBJECT (msg));

 gtk_widget_show_all (msg);
}


static void
menu_export_callback( GtkWidget *w, gpointer op)
{
  GtkWidget *fs = NULL;

  switch( GPOINTER_TO_INT(op))
    {
    case   MENU_OP_EXPORT_HTML:
      fs = gtk_file_selection_new (_("Export HTML"));
      gtk_file_selection_complete (GTK_FILE_SELECTION(fs), "*.html");
      break;
    case   MENU_OP_EXPORT_TEXT:
      fs = gtk_file_selection_new (_("Export Text"));
      break;
    case   MENU_OP_EXPORT_DOCBOOK:
      fs = gtk_file_selection_new (_("Export Docbook"));
      break;

    }

  if (gtk_dialog_run (GTK_DIALOG (fs)) == GTK_RESPONSE_OK)
	{
	  const gchar *selected_filename = 
	    gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));

	  if (selected_filename)
	    switch( GPOINTER_TO_INT(op))
	      {
	      case   MENU_OP_EXPORT_HTML:
		export_html((char *)selected_filename);
		break;
	      case   MENU_OP_EXPORT_TEXT:
		export_text((char *)selected_filename);
		break;
	      case   MENU_OP_EXPORT_DOCBOOK:
		export_docbook((char *)selected_filename);
		break;

	      }
	}

  gtk_widget_destroy (fs);
}

static void 
export_options_dialog(GtkWidget *w,
		      gpointer   data)
{
  GtkWidget *dialog, *bold_checked_button, *color_checked_button,
    *remember_checked_button;
  gint result;
   
  dialog = gtk_dialog_new_with_buttons (_("Export Options"),
					GTK_WINDOW(Fig.window),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_CANCEL,
					GTK_RESPONSE_REJECT,
					GTK_STOCK_OK,
					GTK_RESPONSE_OK,
					NULL);

  remember_checked_button 
    = gtk_check_button_new_with_label(_("Remember last opened"));

  bold_checked_button 
    = gtk_check_button_new_with_label(_("Bold Top Level Items"));

  color_checked_button 
    = gtk_check_button_new_with_label(_("Color Items"));

  /*
  g_signal_connect_swapped (GTK_OBJECT (dialog), 
			    "response", 
			    G_CALLBACK (gtk_widget_destroy),
			    GTK_OBJECT (dialog));
  */
  /* Add the label, and show everything we've added to the dialog. */

  gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
		     remember_checked_button );
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(remember_checked_button),
			       Fig.open_last);

  gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
		     bold_checked_button );
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bold_checked_button),
			       Fig.render_toplevel_pretty);


  gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
		     color_checked_button );
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(color_checked_button),
			       Fig.render_colors);


  gtk_widget_show_all (dialog);

  result = gtk_dialog_run (GTK_DIALOG (dialog));

  switch (result)
    {
    case GTK_RESPONSE_OK:
       break;
    default:
      break;
    }
  gtk_widget_destroy (dialog);

}




/* -- File operations -- */

void 
fig_show_error()
{
  GtkWidget *errbox;
  if (!Fig.lasterror) return;

 errbox = gtk_message_dialog_new(GTK_WINDOW(Fig.window),
				 GTK_DIALOG_DESTROY_WITH_PARENT,
				 GTK_MESSAGE_ERROR,
				 GTK_BUTTONS_OK,
				 Fig.lasterror);

 g_signal_connect_swapped (GTK_OBJECT (errbox), 
			   "response", 
			   G_CALLBACK (gtk_widget_destroy),
			   GTK_OBJECT (errbox));

 gtk_widget_show_all (errbox);

 free(Fig.lasterror);
 Fig.lasterror = NULL;
}

void 
update_disk_sync(gboolean state)
{
  Fig.disk_sync = state;
}

void 
update_filename(gchar *filename)
{
  gchar win_title[256];
  gchar *lastslash;

  if (filename != NULL) /* If NULL passed assume Fig.filename populated */
    {
      if (Fig.filename) free(Fig.filename);
      Fig.filename = (gchar *)strdup((char *)filename);
    }
  else
    {
      filename = Fig.filename;
    }

  lastslash = rindex(filename, '/');
  if (lastslash)
    sprintf(win_title, _("Figment :: %s"), lastslash+1 );
  else
    sprintf(win_title, _("Figment :: %s"), filename);

  gtk_window_set_title (GTK_WINDOW (Fig.window), win_title);

}

void file_save_as (GtkWidget *w, GtkFileSelection *fs) {
  const gchar *selected_filename;

  selected_filename = 
    gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));

  if (selected_filename)
    {
      opml_save((gchar *)selected_filename);
      update_filename((gchar *)selected_filename);
      update_disk_sync(TRUE);
    }
}

void file_open (GtkWidget *w, GtkFileSelection *fs ) {

  const gchar *selected_filename;

  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));

  selected_filename = 
    gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));

  if (selected_filename)
    {
      if (!opml_load((gchar *)selected_filename, GTK_TREE_STORE(model)))
	  fig_show_error();
      else {
	update_filename((gchar *)selected_filename);
	update_disk_sync(TRUE);
      }
    }
}

static void 
menu_file_callback( GtkWidget *w, gpointer op)
{
  GtkWidget *fs = NULL;
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));

  switch( GPOINTER_TO_INT(op))
    {
    case MENU_OP_FILE_SAVE:
      if (Fig.filename != NULL)
	{
	  opml_save(Fig.filename);
	  update_disk_sync(TRUE);
	  return;
	}
    case MENU_OP_FILE_SAVE_AS:
      fs = gtk_file_selection_new (_("Please select new filename"));
      g_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (fs)->ok_button),
			"clicked",
			G_CALLBACK (file_save_as),
			(gpointer)fs);
      break;
    case MENU_OP_FILE_NEW:
      if (!Fig.disk_sync && !fig_question(_("Lose unsaved work?")))
	return;

      if (Fig.filename) free(Fig.filename);
      Fig.filename = NULL;
      gtk_window_set_title (GTK_WINDOW (Fig.window), 
			    _("Figment :: Untitled.xpm"));
      gtk_tree_store_clear (GTK_TREE_STORE(model));
      update_disk_sync(TRUE);
      return;
      break;
    case MENU_OP_FILE_OPEN:
      if (!Fig.disk_sync && !fig_question(_("Lose unsaved work?")))
	return;

      fs = gtk_file_selection_new (_("Please select file to open"));
      g_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (fs)->ok_button),
			"clicked",
			G_CALLBACK (file_open),
			(gpointer) fs);

      break;
    case MENU_OP_FILE_REVERT:
      if (Fig.disk_sync) return; /* Nothing to revert to */
      
      if (fig_question(_("Revert for sure?")))
	{
	  if (!opml_load((gchar *)Fig.filename, GTK_TREE_STORE(model)))
	    fig_show_error();
	  else
	    update_disk_sync(TRUE);
	}

      return;
    }


  gtk_file_selection_complete (GTK_FILE_SELECTION(fs), "*.opml");

  g_signal_connect_swapped (GTK_OBJECT (GTK_FILE_SELECTION(fs)->ok_button),
			    "clicked",
			    G_CALLBACK (gtk_widget_destroy), 
			    (gpointer) fs); 

  g_signal_connect_swapped (GTK_OBJECT (GTK_FILE_SELECTION(fs)->cancel_button),
			    "clicked",
			    G_CALLBACK (gtk_widget_destroy),
			    (gpointer) fs); 

  gtk_widget_show (fs);

}

static void 
menu_view_callback( GtkWidget *w, gpointer op)
{
  GtkTreeIter iter;
  GtkTreePath *path = NULL;
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));
  GtkTreeSelection *selection 
    = gtk_tree_view_get_selection (GTK_TREE_VIEW(Fig.treeview));

  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
    path = gtk_tree_model_get_path(model, &iter);

  switch(GPOINTER_TO_INT (op))
    {
    case MENU_OP_EXPAND:
      if (path)
	gtk_tree_view_expand_row(GTK_TREE_VIEW(Fig.treeview), path, FALSE);
      break;
    case MENU_OP_EXPAND_SUBS:
      if (path)
	gtk_tree_view_expand_row(GTK_TREE_VIEW(Fig.treeview), path, TRUE);
      break;
    case MENU_OP_EXPAND_ALL:
      gtk_tree_view_expand_all(GTK_TREE_VIEW(Fig.treeview));
      break;
    case MENU_OP_COLLAPSE_ALL:
      gtk_tree_view_collapse_all(GTK_TREE_VIEW(Fig.treeview));
      break;
    case MENU_OP_COLLAPSE_PAR:
      if (path && gtk_tree_path_up(path))
	{
	  gtk_tree_view_collapse_row(GTK_TREE_VIEW(Fig.treeview), path);
	  gtk_tree_selection_select_path  (selection, path);
	}
      break;
    case MENU_OP_COLLAPSE:
      if (path)
	gtk_tree_view_collapse_row(GTK_TREE_VIEW(Fig.treeview), path);
      break;
    case MENU_OP_MOVE_NEXT_CHLD:
      if (path) 
	{
	  gtk_tree_path_down(path);
	  gtk_tree_selection_select_path  (selection, path);
	}
      break;
    case MENU_OP_MOVE_NEXT_SIB:
      if (path)
	{
	  gtk_tree_path_next(path);
	  gtk_tree_selection_select_path  (selection, path);
	}
      break;
    case MENU_OP_MOVE_PREV_CHLD:
      if (path && gtk_tree_path_up(path))
	{
	  gtk_tree_selection_select_path  (selection, path);
	}
      break;
    case MENU_OP_MOVE_PREV_SIB:
      if (path && gtk_tree_path_prev(path))
	{
	  gtk_tree_selection_select_path  (selection, path);
	}
      break;
    }

  if (path) gtk_tree_path_free(path);
}

static void 
menu_add_callback( GtkWidget *w, gpointer op )
{
  GtkTreeIter iter;
  GtkTreeIter iter_parent, iter_new, *iter_copy;

  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));
  GtkTreeSelection *selection 
    = gtk_tree_view_get_selection (GTK_TREE_VIEW(Fig.treeview));

  GtkTreePath *new_path = NULL, *path, *old_path;
  GtkTreeViewColumn* col;

  BranchCopy* branch;

      /*      GdkColor color; */

      switch(GPOINTER_TO_INT (op))
	{
	case MENU_OP_ADD_CHLD:
	  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
	    {
	      gtk_tree_store_append (GTK_TREE_STORE(model), 
				     &iter_new, &iter);

	      gtk_tree_store_set (GTK_TREE_STORE(model), &iter_new,
				  CHECKED_COLUMN, FALSE,
				  EDITABLE_COLUMN, TRUE,
				  -1);


	      figment_set_cell_text(&iter_new, _("new item"));

	      update_disk_sync(FALSE);

	      new_path = gtk_tree_model_get_path(model, &iter_new);
	      old_path = gtk_tree_model_get_path(model, &iter);

	      gtk_tree_view_expand_row( GTK_TREE_VIEW(Fig.treeview), 
					old_path, TRUE);
	    }
	  break;
	case MENU_OP_ADD_SIB:
	  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
	    {
	      gtk_tree_store_insert_after(GTK_TREE_STORE(model),
					  &iter_new, NULL, &iter );
	    } else {
	      gtk_tree_store_insert_after(GTK_TREE_STORE(model),
					  &iter_new, NULL, NULL );
	    }

	  gtk_tree_store_set (GTK_TREE_STORE(model), &iter_new,
			      CHECKED_COLUMN, FALSE,
			      EDITABLE_COLUMN, TRUE,
			      -1);

	  update_disk_sync(FALSE);
	  figment_set_cell_text(&iter_new, _("new item"));

	  new_path = gtk_tree_model_get_path(model, &iter_new);
	  break;
	case   MENU_OP_PROMOTE:
	  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
	    {
	      path = gtk_tree_model_get_path(model, &iter);
	      if (gtk_tree_path_get_depth(path) == 1)
		break;


	      if (path && gtk_tree_path_up(path)
		  && gtk_tree_model_get_iter(model, &iter_parent, path))
		{
		  iter_copy = gtk_tree_iter_copy(&iter);
		  branch = branch_recurse_copy(iter_copy, BRANCH_COPY_KIDS);

		  branch_add(&iter_parent, branch, 
			     BRANCH_COPY_ADD_AS_SIBLING_BEFORE, FALSE) ;
		  gtk_tree_store_remove (GTK_TREE_STORE(model), &iter);
		  
		  gtk_tree_iter_free(iter_copy);
		  branch_free(branch);

		  new_path = path;
		  update_disk_sync(FALSE);
		}
	    }
	  break;
	case   MENU_OP_DEMOTE:
	  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
	    {
	      path = gtk_tree_model_get_path(model, &iter);

	      gtk_tree_path_next(path);
	      if (path && gtk_tree_model_get_iter(model, &iter_parent, path))
		{
		  iter_copy = gtk_tree_iter_copy(&iter);
		  branch = branch_recurse_copy(iter_copy, BRANCH_COPY_KIDS);

		  branch_add(&iter_parent, branch, 
			     BRANCH_COPY_ADD_AS_CHILD, TRUE) ;
		  gtk_tree_store_remove (GTK_TREE_STORE(model), &iter);
		  
		  gtk_tree_iter_free(iter_copy);
		  branch_free(branch);

		  new_path = gtk_tree_model_get_path(model, &iter_parent);
		  gtk_tree_path_down(new_path);

		  update_disk_sync(FALSE);
		}

	    }
	  break;
	case   MENU_OP_RAISE:
	  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
	    {
	      path = gtk_tree_model_get_path(model, &iter);
	      
	      if (path && gtk_tree_path_prev(path)  
		  && gtk_tree_model_get_iter(model, &iter_parent, path))
		{
		  iter_copy = gtk_tree_iter_copy(&iter);
		  branch = branch_recurse_copy(iter_copy, BRANCH_COPY_KIDS);

		  branch_add(&iter_parent, branch, 
			     BRANCH_COPY_ADD_AS_SIBLING_BEFORE, FALSE) ;

		  gtk_tree_store_remove (GTK_TREE_STORE(model), &iter);

		  gtk_tree_iter_free(iter_copy);
		  branch_free(branch);
		  update_disk_sync(FALSE);

		  new_path = path;
		}
	    }
	  break;
	case   MENU_OP_LOWER:
	  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
	    {
	      path = gtk_tree_model_get_path(model, &iter);
	      gtk_tree_path_next(path);
	      if (path && gtk_tree_model_get_iter(model, &iter_parent, path))
		{
		  iter_copy = gtk_tree_iter_copy(&iter);
		  branch = branch_recurse_copy(iter_copy, BRANCH_COPY_KIDS);
		  gtk_tree_iter_free(iter_copy);

		  branch_add(&iter_parent, branch, 
			     BRANCH_COPY_ADD_AS_SIBLING_AFTER, FALSE);

		  gtk_tree_store_remove (GTK_TREE_STORE(model), &iter);	
		  branch_free(branch);

		  new_path = path;
		  update_disk_sync(FALSE);
		}
	    }
	  break;

	}

      if (new_path)
	{
	  col = gtk_tree_view_get_column(GTK_TREE_VIEW(Fig.treeview), 
					 THOUGHT_COLUMN);
	  gtk_tree_view_set_cursor (GTK_TREE_VIEW(Fig.treeview),
				    new_path, col, TRUE);
	  gtk_tree_view_scroll_to_cell    (GTK_TREE_VIEW (Fig.treeview),
					   new_path, NULL, TRUE, 0.5, 0.5);

	  gtk_tree_path_free(new_path);
	}
} 

static void remove_item( GtkWidget *w,
                         gpointer   data )
{
  GtkTreeIter iter;
  
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));
  GtkTreeSelection *selection 
    = gtk_tree_view_get_selection (GTK_TREE_VIEW(Fig.treeview));

  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
    {
      GtkWidget *dialog;

      dialog = gtk_message_dialog_new (GTK_WINDOW(Fig.window),
				       GTK_DIALOG_DESTROY_WITH_PARENT,
				       GTK_MESSAGE_QUESTION,
				       GTK_BUTTONS_OK_CANCEL,
				       _("Are you sure?") );

      if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK)
	{
	  gtk_tree_store_remove (GTK_TREE_STORE(model), &iter);
	}

      gtk_widget_destroy (dialog);
      update_disk_sync(FALSE);
    }
}


/* Returns a menubar widget made from the above menu */
GtkWidget *get_menubar_menu( GtkWidget  *window)
{
  GtkItemFactory *item_factory;
  GtkAccelGroup *accel_group;

  accel_group = gtk_accel_group_new ();

  item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>",
                                       accel_group);
  gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);
  gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);

  return gtk_item_factory_get_widget (item_factory, "<main>");
}

static void
item_toggled (GtkCellRendererToggle *cell,
	      gchar                 *path_str,
	      gpointer               data)
{
  GtkTreeModel *model = (GtkTreeModel *)data;
  GtkTreePath *path = gtk_tree_path_new_from_string (path_str);
  GtkTreeIter iter;
  gboolean toggle_item;

  gint *column;

  column = g_object_get_data (G_OBJECT (cell), "column");

  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_model_get (model, &iter, column, &toggle_item, -1);
  toggle_item ^= 1;
  gtk_tree_store_set (GTK_TREE_STORE (model), &iter, column,
		      toggle_item, -1);
  gtk_tree_path_free (path);

  update_disk_sync(FALSE);
}

void
figment_set_cell_text(GtkTreeIter *iter, gchar *text)
{
  const gchar depth_cols[][8] = {
    "black",
    "black",
    "blue",
    "red",
    "green",
    "orange",
    "yellow",
  };

  int depth_cols_idx = 0;
  gchar *marked_up_text = NULL;
  GtkTreePath *path;
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));


  path = gtk_tree_model_get_path(model, iter);


  if (Fig.render_toplevel_pretty || Fig.render_colors)
    {
      marked_up_text = (gchar *)malloc(sizeof(gchar)*(strlen(text)+60));
      memset(marked_up_text, 0, (strlen(text)+60));
      strcat((char *)marked_up_text, "<span ");
      
      if (Fig.render_toplevel_pretty 
	  && gtk_tree_path_get_depth(path) == 1)
	strcat((char *)marked_up_text, "weight='bold' ");
      
      if (Fig.render_colors)
	{
	  depth_cols_idx = gtk_tree_path_get_depth(path);
	  if (depth_cols_idx > 6) depth_cols_idx = 6;
	  strcat((char *)marked_up_text, "color='");
	  strcat((char *)marked_up_text, 
		 depth_cols[depth_cols_idx]);
	  strcat((char *)marked_up_text, "'");
	}

      strcat((char *)marked_up_text, ">");
      strcat((char *)marked_up_text, text);
      strcat((char *)marked_up_text, "</span>");
      
    }
  else
    {
      marked_up_text = g_strdup(text);
    }

  gtk_tree_store_set (GTK_TREE_STORE (model), iter, THOUGHT_COLUMN,
		      text, -1);
     
  gtk_tree_store_set (GTK_TREE_STORE (model), iter, MARKUP_COLUMN,
		      marked_up_text, -1);
  gtk_tree_path_free(path);

}

void
figment_update_cells_prettyness(GtkTreeIter *iter)
{
  gboolean valid = 1;
  GtkTreeModel *model;

  model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));
  
  while (valid)
    {
      gchar *text;
      GtkTreeIter child_iter;
      
      gtk_tree_model_get (model, iter, 
                          THOUGHT_COLUMN, &text,
                          -1);

      figment_set_cell_text(iter, text);

      if (gtk_tree_model_iter_children(model, &child_iter, iter))
	figment_update_cells_prettyness(&child_iter);
      
      valid = gtk_tree_model_iter_next (model, iter);
    }
}


static void
cell_edited (GtkCellRendererText *cell,
	     const gchar         *path_string,
	     const gchar         *new_text,
	     gpointer             data)
{
  GtkTreeModel *model = (GtkTreeModel *)data;
  GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
  GtkTreeIter iter;
  gint *column;

  column = g_object_get_data (G_OBJECT (cell), "column");

  gtk_tree_model_get_iter (model, &iter, path);

  if (GPOINTER_TO_INT (column) == MARKUP_COLUMN)
    {
      gchar *old_text;

      gtk_tree_model_get (model, &iter, MARKUP_COLUMN, &old_text, -1);
      g_free (old_text);

      gtk_tree_model_get (model, &iter, THOUGHT_COLUMN, &old_text, -1);
      g_free (old_text);

      figment_set_cell_text(&iter, (gchar *)new_text);
    }

  update_disk_sync(FALSE);

  gtk_tree_path_free (path);
}

static void
toggle_toolbar(GtkWidget *w,
	       gpointer   data)
{
  if (Fig.show_toolbar)
    {
      gtk_widget_hide(Fig.toolbar);
      Fig.show_toolbar = FALSE;
    } else {
      gtk_widget_show(Fig.toolbar);
      Fig.show_toolbar = TRUE;
    }
}

static void
add_columns (GtkTreeView *treeview)
{
  gint col_offset;
  GtkCellRenderer *renderer, *toggle_renderer;
  GtkTreeViewColumn *column;
  GtkTreeModel *model = gtk_tree_view_get_model (treeview);


  renderer = gtk_cell_renderer_text_new ();
  g_signal_connect (G_OBJECT (renderer), "edited",
		    G_CALLBACK (cell_edited), model);

  g_object_set (G_OBJECT (renderer), "xalign", 0.0, NULL);
  g_object_set_data (G_OBJECT (renderer), "column", (gint *)0);

  col_offset 
    = gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (treeview),
						   -1, "thought",
						   renderer, "markup",
						   MARKUP_COLUMN, "editable" , 
						   EDITABLE_COLUMN, 
						   NULL);
  column = gtk_tree_view_get_column (GTK_TREE_VIEW (treeview), col_offset - 1);
  gtk_tree_view_column_set_clickable (GTK_TREE_VIEW_COLUMN (column), TRUE);
  gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
				   GTK_TREE_VIEW_COLUMN_GROW_ONLY);
  gtk_tree_view_column_set_min_width (GTK_TREE_VIEW_COLUMN (column), 200); 
  gtk_tree_view_column_set_clickable (GTK_TREE_VIEW_COLUMN (column), TRUE);


  toggle_renderer = gtk_cell_renderer_toggle_new ();
  g_object_set (G_OBJECT (toggle_renderer), "xalign", 0.0, NULL);
  g_object_set_data (G_OBJECT (toggle_renderer), "column", 
		     (gint *)CHECKED_COLUMN);

  g_signal_connect (G_OBJECT (toggle_renderer), "toggled", 
		    G_CALLBACK (item_toggled), model);

  col_offset 
    = gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (treeview),
						   -1, "checked",
						   toggle_renderer,
						   "active",
						   CHECKED_COLUMN,
						   NULL);


  column = gtk_tree_view_get_column (GTK_TREE_VIEW (treeview), col_offset - 1);
  gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
  			   GTK_TREE_VIEW_COLUMN_FIXED);
   gtk_tree_view_column_set_fixed_width (GTK_TREE_VIEW_COLUMN (column), 16); 
  gtk_tree_view_column_set_clickable (GTK_TREE_VIEW_COLUMN (column), TRUE);
}

static gint 
win_configure_callback( GtkWidget         *widget, 
			GdkEventConfigure *event,
			gpointer           data )
{
  GtkTreeIter iter;
  GtkTreePath* path;

  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));

  GtkTreeSelection *selection 
    = gtk_tree_view_get_selection (GTK_TREE_VIEW(Fig.treeview));

  /* On matchbox the keyboard will cause the tree to resize, we want to 
     scroll to the current selection                                    */
  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
    {
      path = gtk_tree_model_get_path(model, &iter);
      gtk_tree_view_scroll_to_cell    (GTK_TREE_VIEW (Fig.treeview),
				       path, NULL, FALSE, 0.5, 0.5);
      gtk_tree_path_free(path);
    }

  return FALSE;
}

static void 
fig_options_dialog(GtkWidget *w,
                   gpointer   data)
{
  GtkWidget *dialog, *bold_checked_button, *color_checked_button,
    *remember_checked_button;
  GtkTreeIter iter;
  gint result;
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW(Fig.treeview));

   
  dialog = gtk_dialog_new_with_buttons (_("Options"),
					GTK_WINDOW(Fig.window),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_CANCEL,
					GTK_RESPONSE_REJECT,
					GTK_STOCK_OK,
					GTK_RESPONSE_OK,
					NULL);

  remember_checked_button 
    = gtk_check_button_new_with_label(_("Remember last opened"));

  bold_checked_button 
    = gtk_check_button_new_with_label(_("Bold Top Level Items"));

  color_checked_button 
    = gtk_check_button_new_with_label(_("Color Items"));

  /*
  g_signal_connect_swapped (GTK_OBJECT (dialog), 
			    "response", 
			    G_CALLBACK (gtk_widget_destroy),
			    GTK_OBJECT (dialog));
  */
  /* Add the label, and show everything we've added to the dialog. */

  gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
		     remember_checked_button );
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(remember_checked_button),
			       Fig.open_last);

  gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
		     bold_checked_button );
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bold_checked_button),
			       Fig.render_toplevel_pretty);


  gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
		     color_checked_button );
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(color_checked_button),
			       Fig.render_colors);


  gtk_widget_show_all (dialog);

  result = gtk_dialog_run (GTK_DIALOG (dialog));

  switch (result)
    {
    case GTK_RESPONSE_OK:
      if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bold_checked_button))
	  != Fig.render_toplevel_pretty
	  || 
	  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(color_checked_button))
	  != Fig.render_colors)
	{
	  Fig.render_toplevel_pretty = 
	    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bold_checked_button)); 
	  Fig.render_colors = 
	    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(color_checked_button));
	  gtk_tree_model_get_iter_first(model, &iter);
	  figment_update_cells_prettyness(&iter);
	}

      Fig.open_last 
	= gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(remember_checked_button));
	save_config();
       break;
    default:
      break;
    }
  gtk_widget_destroy (dialog);

}

static int
fig_question(gchar *text)
{
  GtkWidget *msg;
  gint result;
  gint answer = 0;

  msg = gtk_message_dialog_new(GTK_WINDOW(Fig.window),
			       GTK_DIALOG_DESTROY_WITH_PARENT,
			       GTK_MESSAGE_QUESTION,
			       GTK_BUTTONS_OK_CANCEL,
			       text);

  gtk_widget_show_all (msg);

  result = gtk_dialog_run (GTK_DIALOG (msg));
  switch (result)
    {
    case GTK_RESPONSE_OK:
      answer =  1;
      break;
    default:
      break;
    }
  gtk_widget_destroy (msg);

  return answer;
}

static void
fig_exit( GtkWidget *w, gpointer   data )
{
  int op = MENU_OP_FILE_SAVE;

  if (Fig.disk_sync || Fig.filename == NULL) { save_config(); gtk_exit(0); }

  if (fig_question(_("Save work before exit?")))
    {
      opml_save(Fig.filename);
      save_config();
      gtk_exit(0);
    }
}

int main(int argc, char **argv)
{
  GtkWidget *vbox;
  GtkWidget *sw;
  GtkWidget *menubar;
  GtkWidget *button_image;
  /* 
     GtkWidget *view;
  GtkTextBuffer *buffer; 
  GtkWidget *vpaned; */
  GtkTreeModel *model = NULL;
  GtkIconFactory* icon_factory;
  GtkIconSet* icon_set;  
  GdkPixbuf* pixbuf, *pixbuf_title;
  GError *gerror = NULL;

  gtk_init (&argc, &argv);

  gtk_set_locale ();
  setlocale (LC_ALL, "");

  bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
  textdomain (PACKAGE);

  Fig.lasterror = NULL;
  Fig.filename = NULL;

  update_disk_sync(TRUE);
  load_config();
  
  /* create window, etc */
  Fig.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (Fig.window), _("Figment: Untitled.opml"));

  pixbuf_title = gdk_pixbuf_new_from_file( DEFAULTSDIR "/pixmaps/figment.png",
					   &gerror);
  if (pixbuf_title)
    gtk_window_set_icon  (GTK_WINDOW (Fig.window), pixbuf_title);

  icon_factory = gtk_icon_factory_new();

  pixbuf = gdk_pixbuf_new_from_file( DEFAULTSDIR "/figment/add_child_16.png", NULL);
  if (pixbuf == NULL) 
    {
      fprintf(stderr, _("failed to load icon image\n"));
      exit(1);
    }

  icon_set = gtk_icon_set_new_from_pixbuf(pixbuf);
  
  gtk_icon_factory_add(icon_factory,
		       "MY_STOCK_ADD_CHILD",
		       icon_set);

  gtk_icon_factory_add_default(icon_factory);

  g_signal_connect (G_OBJECT (Fig.window), "destroy",
		    G_CALLBACK (gtk_main_quit), NULL);
  g_signal_connect_swapped (G_OBJECT (Fig.window), "delete_event",
			    G_CALLBACK (fig_exit), &Fig.window);
		    
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 0);
  gtk_container_add (GTK_CONTAINER (Fig.window), vbox);

  /* menu bar */
  menubar = get_menubar_menu (Fig.window);
  gtk_box_pack_start (GTK_BOX (vbox), menubar, FALSE, TRUE, 0);

  Fig.toolbar = gtk_toolbar_new();

  gtk_toolbar_set_icon_size(GTK_TOOLBAR(Fig.toolbar), GTK_ICON_SIZE_MENU);
  gtk_toolbar_set_style(GTK_TOOLBAR(Fig.toolbar), GTK_TOOLBAR_ICONS );

  gtk_toolbar_insert_stock(GTK_TOOLBAR(Fig.toolbar), GTK_STOCK_OPEN, 
			   NULL, NULL,
			   G_CALLBACK(menu_file_callback), 
			   (gpointer)MENU_OP_FILE_OPEN, -1);

  gtk_toolbar_insert_stock(GTK_TOOLBAR(Fig.toolbar), GTK_STOCK_SAVE, 
			   NULL, NULL,
			   G_CALLBACK(menu_file_callback), 
			   (gpointer)MENU_OP_FILE_SAVE, -1);

  gtk_toolbar_insert_stock(GTK_TOOLBAR(Fig.toolbar), GTK_STOCK_ADD, 
			   NULL, NULL,
			   G_CALLBACK(menu_add_callback), 
			   (gpointer)MENU_OP_ADD_SIB, -1);

  button_image = gtk_image_new_from_pixbuf(pixbuf);

  gtk_toolbar_append_item (GTK_TOOLBAR(Fig.toolbar),
			   NULL,              
			   NULL,
			   NULL,
			   button_image,
			   G_CALLBACK(menu_add_callback), 
			   (gpointer)MENU_OP_ADD_CHLD);

  gtk_toolbar_insert_stock(GTK_TOOLBAR(Fig.toolbar), GTK_STOCK_DELETE, 
			   NULL, NULL,
			   G_CALLBACK(remove_item), 
			   NULL, -1);

  gtk_toolbar_insert_stock(GTK_TOOLBAR(Fig.toolbar), GTK_STOCK_GO_BACK, 
			   NULL, NULL,
			   G_CALLBACK(menu_add_callback), 
			   (gpointer)MENU_OP_PROMOTE, -1);

  gtk_toolbar_insert_stock(GTK_TOOLBAR(Fig.toolbar), GTK_STOCK_GO_FORWARD, 
			   NULL, NULL,
			   G_CALLBACK(menu_add_callback), 
			   (gpointer)MENU_OP_DEMOTE, -1);

  gtk_toolbar_insert_stock(GTK_TOOLBAR(Fig.toolbar), GTK_STOCK_GO_UP, 
			   NULL, NULL,
			   G_CALLBACK(menu_add_callback), 
			   (gpointer)MENU_OP_RAISE, -1);

  gtk_toolbar_insert_stock(GTK_TOOLBAR(Fig.toolbar), GTK_STOCK_GO_DOWN, 
			   NULL, NULL,
			   G_CALLBACK(menu_add_callback), 
			   (gpointer)MENU_OP_LOWER, -1);

  gtk_box_pack_start (GTK_BOX (vbox), Fig.toolbar, FALSE, TRUE, 0);

  
  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_set_border_width (GTK_CONTAINER (sw), 4);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
				       GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start (GTK_BOX (vbox), sw, TRUE, TRUE, 0);

  model = GTK_TREE_MODEL (gtk_tree_store_new (NUM_COLUMNS,
					      G_TYPE_STRING,
					      G_TYPE_BOOLEAN,
					      G_TYPE_STRING,
					      G_TYPE_BOOLEAN
					      ));
  
  /* create tree view */
  Fig.treeview = gtk_tree_view_new_with_model(model);

  g_object_unref (G_OBJECT (model));
  gtk_tree_view_set_headers_visible(GTK_TREE_VIEW (Fig.treeview), FALSE);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (Fig.treeview), FALSE);
  gtk_tree_view_set_reorderable (GTK_TREE_VIEW (Fig.treeview), FALSE);

  gtk_tree_selection_set_mode (gtk_tree_view_get_selection (GTK_TREE_VIEW (Fig.treeview)),
			       GTK_SELECTION_SINGLE);
  
  add_columns (GTK_TREE_VIEW (Fig.treeview));
  
  gtk_container_add (GTK_CONTAINER (sw), Fig.treeview);

  g_signal_connect(G_OBJECT (Fig.window), "configure_event",
		    G_CALLBACK (win_configure_callback), NULL);

  gtk_window_set_default_size (GTK_WINDOW (Fig.window), 650, 400);

  gtk_widget_show_all (Fig.window);

  if (!Fig.show_toolbar)
      gtk_widget_hide(Fig.toolbar);

  if (argv[1]) 
    {
      Fig.open_last = TRUE;
      Fig.filename = argv[1];
    }

  if (Fig.open_last)
    {
      opml_load((gchar *)Fig.filename, GTK_TREE_STORE(model));
      update_filename((gchar *)NULL);
      update_disk_sync(TRUE);
    }

  gtk_main();

  return 0;
}
